#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Standardize methods for file handling. 
    The functions include verifing filenames are safe for OS in use.
    Files are handled using the pathlib approach.
    Sub-Directory structures can be generated from timestamps.
    Sub-Directory structures can be generated from filenames.
"""

from os import error
from os import chdir as Change_Directory
from loguru import logger
from pathlib import Path
import shutil as _shutil
from shutil import SpecialFileError
from cfsiv_utils.time_strings import timefstring



def get_files(source_directory: Path, pattern=None):
    """Returns a list of all files from directory matching optional pattern.

    Args:
        source_directory (Path): Relative or absolute reference to a directory.
        pattern (str, optional): Filename matching pattern wildcards accepted. Defaults to '*.*'.

    Raises:
        TypeError: If optional inputs are out of bounds.

    Returns:
        list: A list of all files matching pattern from directory.
    """
    if pattern == None:
        pattern = '*.*'
    else:
        if type(pattern) != str:
            raise TypeError(f'pattern must be type string, got {type(pattern)}')
    if type(source_directory) != Path:
        logger.debug(f'source_directory must be type Path. Got: {type(source_directory)}')
        source_directory = Path(source_directory)        
    ORIGINAL_WORKING_DIRECTORY = Path.cwd()
    SOURCE_DIRECTORY = source_directory.resolve()
    if not(SOURCE_DIRECTORY.is_dir()):
        logger.error(f'Source_directory must be a valid dir. Got {source_directory}')
    NEW_WORKING_DIRECTORY = Path.cwd()    
    if SOURCE_DIRECTORY != ORIGINAL_WORKING_DIRECTORY:
        try:
            Change_Directory(SOURCE_DIRECTORY)
            NEW_WORKING_DIRECTORY = Path.cwd()
        except error as e:
            logger.error(f'Could not access source directory: {e}')
    if SOURCE_DIRECTORY != NEW_WORKING_DIRECTORY:
        logger.error('Unknown change directory to source error.')
        raise FileNotFoundError(f'Could not access source_Path: {source_directory}')
    try:
        files = list(SOURCE_DIRECTORY.rglob(pattern))
    except error as e:
        logger.error(f'Error locating files matching pattern: {pattern}')
        files = []
    # TODO return cwd to original location.
    try:
        Change_Directory(ORIGINAL_WORKING_DIRECTORY)
    except error as e:
        logger.error(f'Could not restore original directory: {e}')    
    return files



def clean_filename_str(fn: str):
    """Replace invalid characters from provided string.
        Note: '-' is invalid in windows if it is the last character in a name.
        # TODO replace invalid characters with underscores.
    """
    return Path("".join(i for i in fn if i not in "\/:*?<>|-"))



def new_name_if_exists(file: Path):
    """Make a new filename that avoids name collisions.
        example: filename(xx).ext where xx is incremented until 
        unused filename is created.

        Args:
            file (Path): proposed unique filename.

        Returns:
            Path_obj: Guaranteed unique filename.
    """
    new_name = file
    i = 1
    while True:
        if not new_name.is_file():
            return new_name
        else:
            new_name = file.with_name(f"{file.stem}({i}){file.suffix}")
            i += 1



def copy_to_target(file: Path, target_diectory=None):
    """Copy offered file to new location
        while ensuring not to overwrite any existing file.

    Args:
        file (Path): Filename to copy to new location.
        target_directory (Path, optional): must be a valid Path if provided.
            Current working directory will be used as default if not provided.

    Returns:
        bool: Always returns True unless 'SpecialFileError' is raised.
    """
    if target_diectory == None:
        target_diectory = Path.cwd()
    else:
        if type(target_diectory) != Path:
            raise FileNotFoundError(f'target_directory must be type Path, got {type(target_diectory)}')
    new_file = new_name_if_exists(target_diectory / file.name)
    try:
        _shutil.copy2(file, new_file)
    except error as e:
        raise SpecialFileError(f"could not copy file '{file}' due to: {e}")
    return True



def copy_to_target_and_divide_by_filedate(file: Path, target_directory=None):
    """Generate a destination for the offered file based on its' timestamp.
        Destination is in the form root/year/month/filename.ext

    Args:
        file (Path): File to copy as a Path_obj
        target_directory (Path, optional): defaults to current working directory.

    Returns:
        bool: Always returns True at this point.
    """
    if target_directory == None:
        target_directory = Path.cwd()
    else:
        if type(target_directory) != Path:
            raise TypeError(f'target directory must be a valid Path, got {type(target_directory)}')
    creation_date = file.stat().st_mtime # in windows this is closer to the oldest date on the file.
    # st_ctime will be equal to the most recent time the file was copied from place to place.
    date = timefstring(creation_date)
    new_path = target_directory / f"{date.year}/{date.month:02}/"
    new_path.mkdir(parents=True, exist_ok=True)
    return copy_to_target(file.name, new_path)



def copy_to_target_and_divide_by_dictionary(file: Path, target_directory=None, characters=None):
    """Generate a destination for the offered file based on its' name.
        Destination is in the form root/(first 'x' characters of filename)/filename.ext

    Args:
        file (Path): File to copy as a Path_obj
        target_directory (Path, optional): root directory for destination.
        characters (integer, optional): number of characters from filename to use.

    Returns:
        bool: status of copy process.
    """
    if characters == None:
        characters = 1 # defaults to the first character of the filename.
    else:
        if type(characters) != int:
            raise(f'characters value must be integer, got {type(characters)}')
    if target_directory == None:
        target_directory = Path.cwd() # defaults to the current working directory.
    else:
        if type(target_directory) != Path:
            raise TypeError(f'target directory must be a valid Path, got {type(target_directory)}')
    new_path = target_directory / f"{file.name[0:characters]}/"
    new_path.mkdir(parents=True, exist_ok=True)
    return copy_to_target(file.name, new_path)



def check_and_validate_fname(fname, target_directory=None, rename=None):
    """Remove invalid characters in filename, combine with target_directory (optional)
    and optionaly create a new filename that doesn't already exist at destination.  

    Args:
        fname (Path): filename to fix.
        target_directory (Path, optional): defaults to current working directory.
        rename (bool, optional): generates a name that is unique. Defaults to False.

    Returns:
        Path: Path object that is valid and does not exist.
    """
    if target_directory == None:
        target_directory = Path.cwd() # defaults to the current working directory.
    else:
        if type(target_directory) != Path:
            raise TypeError(f'target directory must be a valid Path, got {type(target_directory)}')
    if rename == None:
        rename = False # default
    else:
        if type(rename) != bool:
            raise TypeError(f'rename must be Bool, got {type(rename)}')            
    clean_name = clean_filename_str(fname)
    target_directory.mkdir(parents=True, exist_ok=True)
    OUT_PATH_HANDLE = Path(target_directory, clean_name)
    if OUT_PATH_HANDLE.is_file():
        if rename:
            unique_name = new_name_if_exists(OUT_PATH_HANDLE)
        else:
            return None
    return unique_name



@logger.catch
def Main():
    try:
        filelist = get_files(Path('Q:'), pattern='*takeout*.zip')
    except FileNotFoundError as e:
        filelist = []
        print(f'Bad Path: {e}')
    print(len(filelist))


    data = ['qwerty~!@#$%^&*().ext', Path('qwerty~!@#$%^&().ext')]
    reslt = clean_filename_str(data[0])
    assert data[1] == reslt

    testname = new_name_if_exists(Path('README.md'))
    assert testname == Path('README(1).md')

    # copy_to_target(

    # copy_to_target_and_divide_by_filedate(

    # copy_to_target_and_divide_by_dictionary(

    # check_and_validate_fname(


if __name__ == '__main__':
    Main()

# NOTES FOR USE OF pathlib

""" access parts of a filename:
>>> Path('static/dist/js/app.min.js').name
'app.min.js'
>>> pathobj = Path('static/dist/js/app.min.js')
>>> print(pathobj.name)
'app.min.js'
>>> Path('static/dist/js/app.min.js').suffix
'.js'
>>> Path('static/dist/js/app.min.js').suffixes
'.min.js'
>>> Path('static/dist/js/app.min.js').stem
'app'
"""

# set your reference point with the location of the python file you’re writing in
# this_file = Path(__file__)

# Here are three ways to get the folder of the current python file
# this_folder1 = Path(__file__, "..")
# this_folder2 = Path(__file__) / ".."
# this_folder3 = Path(__file__).parent

# This will fail:
# assert this_folder1 == this_folder2 == this_folder3
# becasue the variables are relative paths.

# The resolve() method removes '..' segments, follows symlinks, and returns
# the absolute path to the item.
# this works:
# assert this_folder1.resolve() == this_folder2.resolve() == this_folder3.resolve()


# folder_where_python_was_run = Path.cwd()

# create a new folder:
# Path("/my/directory").mkdir(parents=True, exist_ok=True)

# find path to your running code
# project_root = Path(__file__).resolve().parents[1] # this is the folder 2 levels up from your running code.

# create a new PathObj:
# static_files = project_root / 'static' # if you like this sort of thing they overrode the divide operator.
# media_files = Path(project_root, 'media') # I prefer this method.

# how to define 2 sub-directories at once:
# compiled_js_folder = static_files.joinpath('dist', 'js') # this is robust across all OS.

# gather items from specified path
# list(static_files.iterdir()) # returns a list of all items in directory.
# [x for x in static_files.iterdir if x.is_dir()] # list of only directories in a folder.
# [x for x in static_files.iterdir if x.is_file()] # same for files only.

# get a list of items matching a pattern:
# files = list(compiled_js_folder.glob('*.js')) # returns files ending with '.js'.

# sort the list of files by timestamp:
# files.sort(key=lambda fn: fn.stat().st_atime)
# atime, mtime, ctime don't seem to mean what I think. (Pathlib error?)

# search recursively down your folders path:
# sorted(project_root.rglob('*.js'))

# verify a path exists:
# Path('relative/path/to/nowhere').exists() # returns: False

# Example of directory deletion by pathlib
# pathobj = Path("demo/")
# pathobj.rmdir()

# Example of file deletion by pathlib
# pathobj = Path("demo/testfile.txt")
# pathobj.unlink()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# version 2.0
"""Standardize time strings and datetime objects used in a project.
"""


from datetime import datetime, date, timezone, timedelta
from time import sleep
import pytz
from loguru import logger


tz_UTC = pytz.timezone("UTC")
tz_LOCAL = pytz.timezone("America/Louisville")


@logger.catch
def timefstring(dtobj, tz_name=True):
    """Standardize the format used for timestamp string format.
    Include 3 letter string for timezone if set to True.
    """
    if tz_name:
        return f'{dtobj.strftime("%Y-%m-%d_%H:%M:%S%Z")}'
    else:
        return f'{dtobj.strftime("%Y-%m-%d_%H:%M:%S")}NTZ' #NTZ = Naive Time Zone


def LOCAL_TODAY():
    return  date.today()

def UTC_NOW():
    return datetime.now(tz_UTC)

def LOCAL_CURRENT_YEAR():
    return str(LOCAL_TODAY().year)

def LOCAL_TODAY_STRING():
    return LOCAL_TODAY().strftime("%Y-%m-%d")

def UTC_NOW_STRING():
    return timefstring(UTC_NOW())

def LOCAL_NOW():
    return datetime.now(tz_LOCAL)

def LOCAL_NOW_STRING():
    return timefstring(LOCAL_NOW())


if __name__ == "__main__":
    for i in range(5):
        print(f"LOCAL TODAY: {LOCAL_TODAY()} type: {type(LOCAL_TODAY())}")
        print(f"UTC NOW: {UTC_NOW()} type: {type(UTC_NOW())}")
        print(f"LOCAL NOW: {LOCAL_NOW()} type: {type(LOCAL_NOW())}")
        print(f"LOCAL CURRENT_YEAR: {LOCAL_CURRENT_YEAR()} type: {type(LOCAL_CURRENT_YEAR())}")
        print(f"LOCAL TODAY_STRING: {LOCAL_TODAY_STRING()} type: {type(LOCAL_TODAY_STRING())}")
        print(f"UTC NOW_STRING: {UTC_NOW_STRING()} type: {type(UTC_NOW_STRING())}")
        print(f"LOCAL NOW_STRING: {LOCAL_NOW_STRING()} type: {type(LOCAL_NOW_STRING())}")
        sleep(2)
        print()


def apply_logical_year_value_to_monthday_pair(datestring, scrape_datestamp):
    """Given a month and day apply the rule that it must represent a day in the near past or future.
    This problem presents itself when gathering data from the National Weather Service.
    The site I am scraping for the readings of river water level values does only include the month/day not year.
    These datapoints are historical going back only approximately 30 days and forecast only 7 days future.
    At the end of a year and the first month of the year they are problematic.

    Args:
        datestring (str): 'any parseable string representing a month and day'
        scrape_datestamp (datetime.datetime):  actual date of web scrape.

    Returns:
        datetime.object: fully qualified date with the corrected year.
    """
    from dateutil.parser import parse, ParserError
    try:
        supplied_date = parse(datestring)
    except ParserError as e:
        print(f'{e}: Could not parse datestring provided.')
        raise ParserError(e)

    # All work is done with timezone aware objects.
    supplied_date = supplied_date.replace(tzinfo=pytz.UTC)
    scrape_datestamp = scrape_datestamp.replace(tzinfo=pytz.UTC)

    # expceted result: type datetime(yyyy, mm, dd, hh, min, sec) object
    sc_yyyy = scrape_datestamp.strftime("%Y")
    su_mnth = supplied_date.strftime("%m")
    su_dy = supplied_date.strftime("%d")

    if len(sc_yyyy)+len(su_mnth)+len(su_dy) != 8 or not(f'{sc_yyyy}{su_mnth}{su_dy}'.isdigit()):
        print(f'Parsed date returned y={sc_yyyy} m={su_mnth} d={su_dy}')
        raise AttributeError('Argument must be a parseable datestring.')

    offered_year = int(sc_yyyy)
    prev_year = offered_year-1
    next_year = offered_year+1

    mm = int(su_mnth)
    dd = int(su_dy)

    # Create a UTC timezone instance
    UTC_TimeDelta = timedelta(hours=0)
    tzUTC = timezone(UTC_TimeDelta, name="UTC")

    # create list of possible dates
    dates = []
    offered_datestamp = datetime(offered_year,mm,dd,1,0,0,0,tzUTC)
    dates.append(offered_datestamp)
    prev_datestamp = datetime(prev_year,mm,dd,1,0,0,0,tzUTC)
    dates.append(prev_datestamp)
    next_datestamp = datetime(next_year,mm,dd,1,0,0,0,tzUTC)
    dates.append(next_datestamp)

    # Create a dict of dates keyed on the number of days elapsed.
    delta_dict = {}
    for _i, date in enumerate(dates):
        delta = scrape_datestamp - date
        # print(f'{scrape_datestamp} scrape, {date} date')
        delta_dict[abs(delta.days)] = date

    # Sort and return only the date nearest in time to year specified.
    keys = sorted(delta_dict.keys())
    corrected_datestamp = delta_dict[keys[0]]

    # print(f'Offered:{datestring} and scrape year {scrape_datestamp}, Corrected:{corrected_datestamp}')    
    # print(f'{delta_dict}')

    return corrected_datestamp




# Notes:
"""
# working with datetime objects:

#    It is allowed to subtract one datetime object from another datetime object
#    The resultant object from subtraction of two datetime objects is an object of type timedelta

import datetime

d1 = datetime.datetime.now()
d2 = datetime.datetime.now()
x  = d2-d1
print(type(x))
print(x)
 
    Output:
        <class 'datetime.timedelta'>
        0:00:00.000008

#    The subtraction operation of two datetime objects succeeds, if both the objects are naive objects, 
#    or both the objects are aware objects. i.e., Both the objects have their tzinfo as None
#    If one datetime object is a naive object and another datetime object is an aware object 
#    then  python raises an exception of type TypeError stating:
#       TypeError: can't subtract offset-naive and offset-aware datetimes.

import datetime
# Time difference for pacific time
pacificTimeDelta    = datetime.timedelta(hours=-8)

# Create a PST timezone in instance
tzPST        = datetime.timezone(pacificTimeDelta, name="PST")

# create a datetime for 1 PM PST - 31Dec2017 - An aware Object
OnePMyearEndPST = datetime.datetime(2017,12,31,1,0,0,0,tzPST)

# create a dateobject for current time - without any timezone attached - A naive object
todaysDate = datetime.datetime.now()

# Try subtracting the naive object from the aware object
    # This will raise a Type Error
todaysDate = OnePMyearEndPST - todaysDate
    TypeError: can't subtract offset-naive and offset-aware datetimes

# Time difference for EST time
nycTimeDelta    = datetime.timedelta(hours=-8)

# Create a PST timezone in instance
tzEST           = datetime.timezone(nycTimeDelta, name="EST")

# create a datetime for 1 PM PST - 31Dec2017 - An aware Object
TwoPMyearEndEST = datetime.datetime(2017,12,31,2,0,0,0,tzEST)

# If both the datetime objects are aware objects and each object is having a different time zone information 
#   – then both the time quantity of the objects are converted into UTC first 
#       and the subtraction is applied on the two UTC time quantities.
print(TwoPMyearEndEST-OnePMyearEndPST)
    Output:
        1:00:00







Dateparser Search
Search for Dates in Longer Chunks of Text

Warning

Support for searching dates is really limited and needs a lot of improvement, we look forward to community’s contribution to get better on that part. See “Contributing”.

You can extract dates from longer strings of text. They are returned as list of tuples with text chunk containing the date and parsed datetime object.

dateparser.search.search_dates(text, languages=None, settings=None, add_detected_language=False)[source]

    Find all substrings of the given string which represent date and/or time and parse them.
    Parameters:	

        text (str) – A string in a natural language which may contain date and/or time expressions.
        languages (list) – A list of two letters language codes.e.g. [‘en’, ‘es’]. If languages are given, it will not attempt to detect the language.
        settings (dict) – Configure customized behavior using settings defined in dateparser.conf.Settings.
        add_detected_language (bool) – Indicates if we want the detected language returned in the tuple.

    Returns:	

    Returns list of tuples containing: substrings representing date and/or time, corresponding datetime.datetime object and detected language if add_detected_language is True. Returns None if no dates that can be parsed are found.
    Return type:	

    list
    Raises:	

    ValueError - Unknown Language

    >>> from dateparser.search import search_dates
    >>> search_dates('The first artificial Earth satellite was launched on 4 October 1957.')
    [('on 4 October 1957', datetime.datetime(1957, 10, 4, 0, 0))]

    >>> search_dates('The first artificial Earth satellite was launched on 4 October 1957.',
    >>>              add_detected_language=True)
    [('on 4 October 1957', datetime.datetime(1957, 10, 4, 0, 0), 'en')]

    >>> search_dates("The client arrived to the office for the first time in March 3rd, 2004 "
    >>>              "and got serviced, after a couple of months, on May 6th 2004, the customer "
    >>>              "returned indicating a defect on the part")
    [('in March 3rd, 2004 and', datetime.datetime(2004, 3, 3, 0, 0)),
     ('on May 6th 2004', datetime.datetime(2004, 5, 6, 0, 0))]

Advanced Usage

If you need more control over what is being parser check the Settings section as well as the Using DateDataParser section.


"""


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# version 2.0
"""Create a file and populate with items from a data structure.
    # TODO move this functionality into the filehandling module and standardize.
"""

import csv
from pathlib import Path
from filehandling import check_and_validate
from loguru import logger


@logger.catch
def write_csv(data, filename="temp.csv", directory="CSV_DATA", use_subs=False):
    """'data' is expected to be a list of dicts 
    Take data and write all fields to storage as csv with headers from keys.
    if filename already exists automatically append to end of file if headers match.

with open(r'names.csv', 'a', newline='') as csvfile:
    fieldnames = ['This','aNew']
    writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
    writer.writerow({'This':'is', 'aNew':'Row'})

    Process: file exist? headers match? append data.
    file exist? headers mis-match. raise exception
    no file? create file and save data.
    """
    # create csv file path
    dirobj = Path(Path.cwd(), directory)
    dirobj.mkdir(parents=True, exist_ok=True)    
    pathobj = check_and_validate(filename, dirobj)

    if not pathobj.exists():
        with open(pathobj, "w", newline="") as csvfile:
            csv_obj = csv.writer(csvfile, delimiter=",")
            headers = data[0].keys()
            csv_obj.writerow(headers)
    with open(pathobj, "a+", newline="") as csvfile:
        csv_obj = csv.writer(csvfile, delimiter=",")
        for item in data:  # data should be the list of dicts contaning observations/forecasts.
            csv_obj.writerow(item.values())

    return



if __name__ == "__main__":
    from pathlib import Path
    this_file = Path(__file__)
    print(f'This file {this_file} has no current standalone function.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" This file contains basic webscraping elements found in 
https://realpython.com/python-web-scraping-practical-introduction/
"""

from requests import get
from requests.exceptions import RequestException
from contextlib import closing
from bs4 import BeautifulSoup

from loguru import logger
from pathlib import Path
from filehandling import check_and_validate
from time_strings import UTC_NOW_STRING


@logger.catch
def simple_get(url):
    """
    Attempts to get the content at `url` by making an HTTP GET request.
    If the content-type of response is some kind of HTML/XML, return the
    text content, otherwise raise exception if bad response or RequestException
    """
    try:
        with closing(get(url, stream=True)) as resp:
            if is_good_response(resp):
                return resp.content
            else:
                raise Warning(f'Bad response from: {url}')

    except RequestException as e:
        log_error(f"Error during requests to {url} : {str(e)}")
        raise Warning(f'{str(e)}')



@logger.catch
def is_good_response(resp):
    """
    Returns True if the response seems to be HTML, False otherwise.
    """
    content_type = resp.headers["Content-Type"].lower()
    return (
        resp.status_code == 200
        and content_type is not None
        and content_type.find("html") > -1
    )


@logger.catch
def log_error(e):
    """
    It is always a good idea to log errors. 
    This function just prints them, but you can
    make it do anything.
    """
    print(e)


def save_html_text(txt):
    """Place 'txt' in a subdirectory in the current working directory 
    under a filename based on the current time.
    This will be useful for additional review of accuracy in data extraction.
    """
    # create file path
    filename = f'{UTC_NOW_STRING()}_webscrape.rawhtml'
    # use a subdirectory of current working directory
    dirobj = Path(Path.cwd(), 'raw_web_scrapes')
    dirobj.mkdir(parents=True, exist_ok=True)   
    # ensure that filename only contains valid characters 
    pathobj = check_and_validate(filename, dirobj)
    with open(pathobj, "w") as txtfile:
        txtfile.write(str(txt))
    return None


@logger.catch
def retrieve_cleaned_html(url, cache=False):
    raw_resp = simple_get(url)
    if raw_resp is not None:
        if cache:
            save_html_text(raw_resp)
        return BeautifulSoup(raw_resp, "html.parser")
    return None
